// This file is part of the Acts project.
//
// Copyright (C) 2017-2018 CERN for the benefit of the Acts project
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#pragma once

#include "Acts/Geometry/GeometryContext.hpp"
#include "Acts/Geometry/ILayerBuilder.hpp"
#include "Acts/Geometry/LayerCreator.hpp"
#include "Acts/Geometry/ProtoLayerHelper.hpp"
#include "Acts/Geometry/SurfaceBinningMatcher.hpp"
#include "Acts/Plugins/TGeo/ITGeoIdentifierProvider.hpp"
#include "Acts/Utilities/BinningType.hpp"
#include "Acts/Utilities/Definitions.hpp"
#include "Acts/Utilities/Logger.hpp"
#include "Acts/Utilities/Units.hpp"

#include <climits>
#include <tuple>

class TGeoMatrix;
class TGeoVolume;
class TGeoNode;

namespace Acts {

class TGeoDetectorElement;
class Surface;

using namespace Acts::UnitLiterals;

/// @class TGeoLayerBuilder
///
/// This parses the gGeoManager and looks for a defined combination
/// of volume with contained sensitive detector element. The association
/// is done by matching the names of the TGeoNode / TGeoVolume to the search
/// string.
///
/// The parsing can be restricted to a given parse volume (in r and z),
/// and given some splitting parameters the surfaces can be automatically be
/// split into layers.
class TGeoLayerBuilder : public ILayerBuilder {
 public:
  ///  Helper config structs for volume parsin
  struct LayerConfig {
   public:
    using RangeConfig = std::pair<BinningValue, std::pair<double, double>>;

    using SplitConfig = std::pair<BinningValue, double>;

    /// Identify the search volume by name
    std::string volumeName = "";
    /// Identify the sensor(s) by name
    std::vector<std::string> sensorNames = {};
    /// The local axis definition of TGeo object to Acts::Surface
    std::string localAxes = "XYZ";
    /// Parse ranges: parameter and ranges
    std::vector<RangeConfig> parseRanges = {};
    /// Layer splitting: parameter and tolerance
    std::vector<SplitConfig> splitConfigs = {};
    /// The envelope to be built around the layer
    std::pair<double, double> envelope = {0_mm, 0_mm};
    /// Binning setup in l0: nbins (-1 -> automated), axis binning type
    std::tuple<int, BinningType> binning0 = {-1, equidistant};
    /// Binning setup in l1: nbins (-1 -> automated), axis binning type
    std::tuple<int, BinningType> binning1 = {-1, equidistant};

    // Default constructor
    LayerConfig()
        : volumeName(""),
          sensorNames({}),
          localAxes("XZY"),
          envelope(std::pair<double, double>(1_mm, 1_mm)) {}
  };

  /// @struct Config
  /// @brief nested configuration struct for steering of the layer builder
  struct Config {
    /// String based identification
    std::string configurationName = "undefined";
    /// Unit conversion
    double unit = 1_cm;
    /// Create an indentifier from TGeoNode
    std::shared_ptr<const ITGeoIdentifierProvider> identifierProvider = nullptr;
    /// Layer creator
    std::shared_ptr<const LayerCreator> layerCreator = nullptr;
    /// ProtoLayer helper
    std::shared_ptr<const ProtoLayerHelper> protoLayerHelper = nullptr;
    /// Configuration is always | n | c | p |
    std::array<std::vector<LayerConfig>, 3> layerConfigurations;
    /// Split tolerances in R
    std::array<double, 3> layerSplitToleranceR = {-1., -1., -1.};
    /// Split tolerances in Z
    std::array<double, 3> layerSplitToleranceZ = {-1., -1., -1.};
    /// Automated binning & tolerances
    bool autoSurfaceBinning = false;
    /// The surface binning matcher
    Acts::SurfaceBinningMatcher surfaceBinMatcher;
  };

  /// Constructor
  /// @param config is the configuration struct
  /// @param logger the local logging instance
  TGeoLayerBuilder(const Config& config,
                   std::unique_ptr<const Logger> logger =
                       getDefaultLogger("TGeoLayerBuilder", Logging::INFO));

  /// Destructor
  ~TGeoLayerBuilder() override;

  /// LayerBuilder interface method - returning the layers at negative side
  ///
  /// @param gctx the geometry context for this build call
  ///
  const LayerVector negativeLayers(const GeometryContext& gctx) const final;

  /// LayerBuilder interface method - returning the central layers
  ///
  /// @param gctx the geometry context for this build call
  ///
  const LayerVector centralLayers(const GeometryContext& gctx) const final;

  /// LayerBuilder interface method - returning the layers at negative side
  ///
  /// @param gctx the geometry context for this build call
  ///
  const LayerVector positiveLayers(const GeometryContext& gctx) const final;

  /// Name identification
  const std::string& identification() const final;

  /// Set the configuration object
  /// @param config is the configuration struct
  void setConfiguration(const Config& config);

  /// Get the configuration object
  Config getConfiguration() const;

  /// Set logging instance
  void setLogger(std::unique_ptr<const Logger> newLogger);

  /// Return the created detector elements
  const std::vector<std::shared_ptr<const TGeoDetectorElement>>&
  detectorElements() const;

 private:
  /// Configuration object
  Config m_cfg;

  /// layer types
  std::array<std::string, 3> m_layerTypes = {"Negative", "Central", "Positive"};

  /// Private access to the logger
  const Logger& logger() const { return *m_logger; }

  /// Logging instance
  std::unique_ptr<const Logger> m_logger;

  /// @todo make clear where the TGeoDetectorElement lives
  std::vector<std::shared_ptr<const TGeoDetectorElement>> m_elementStore;

  /// Private helper method : build layers
  ///
  /// @param gcts the geometry context of this call
  /// @param layers is goint to be filled
  /// @param type is the indication which ones to build -1 | 0 | 1
  void buildLayers(const GeometryContext& gctx, LayerVector& layers,
                   int type = 0);

  /// Private helper method : register splitting input
  void registerSplit(std::vector<double>& parameters, double test,
                     double tolerance, std::pair<double, double>& range) const;
};

inline void TGeoLayerBuilder::registerSplit(
    std::vector<double>& parameters, double test, double tolerance,
    std::pair<double, double>& range) const {
  bool found = false;
  // min/max setting
  range.first = std::min(range.first, test);
  range.second = std::max(range.second, test);
  // Loop and find the split parameters
  for (auto& splitPar : parameters) {
    if (std::abs(test - splitPar) < tolerance) {
      found = true;
    }
  }
  if (not found) {
    parameters.push_back(test);
  }
}

inline TGeoLayerBuilder::Config TGeoLayerBuilder::getConfiguration() const {
  return m_cfg;
}

inline const std::vector<std::shared_ptr<const TGeoDetectorElement>>&
TGeoLayerBuilder::detectorElements() const {
  return m_elementStore;
}

inline const std::string& TGeoLayerBuilder::identification() const {
  return m_cfg.configurationName;
}

}  // namespace Acts
