// This file is part of the Acts project.
//
// Copyright (C) 2019 CERN for the benefit of the Acts project
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#include <boost/test/data/test_case.hpp>
#include <boost/test/unit_test.hpp>

#include "Acts/EventData/NeutralTrackParameters.hpp"
#include "Acts/EventData/TrackParameters.hpp"
#include "Acts/Geometry/GeometryContext.hpp"
#include "Acts/Tests/CommonHelpers/FloatComparisons.hpp"
#include "Acts/Utilities/Definitions.hpp"
#include "Acts/Utilities/UnitVectors.hpp"
#include "Acts/Utilities/Units.hpp"

#include <limits>

#include "TrackParametersDatasets.hpp"

namespace {

using namespace Acts;
using namespace Acts::UnitLiterals;
using AnyFreeTrackParameters = SingleFreeTrackParameters<AnyCharge>;

constexpr auto eps = 8 * std::numeric_limits<FreeScalar>::epsilon();
const GeometryContext geoCtx;
const FreeSymMatrix cov = FreeSymMatrix::Identity();

template <typename charge_t>
void checkParameters(const SingleFreeTrackParameters<charge_t>& params,
                     const Vector4D& pos4, const Vector3D& unitDir, double p,
                     double q) {
  const auto qOverP = (q != 0) ? (q / p) : (1 / p);
  const auto pos = pos4.segment<3>(ePos0);

  // native values
  CHECK_CLOSE_OR_SMALL(params.template get<eFreePos0>(), pos4[ePos0], eps, eps);
  CHECK_CLOSE_OR_SMALL(params.template get<eFreePos1>(), pos4[ePos1], eps, eps);
  CHECK_CLOSE_OR_SMALL(params.template get<eFreePos2>(), pos4[ePos2], eps, eps);
  CHECK_CLOSE_OR_SMALL(params.template get<eFreeTime>(), pos4[eTime], eps, eps);
  CHECK_CLOSE_OR_SMALL(params.template get<eFreeDir0>(), unitDir[eMom0], eps,
                       eps);
  CHECK_CLOSE_OR_SMALL(params.template get<eFreeDir1>(), unitDir[eMom1], eps,
                       eps);
  CHECK_CLOSE_OR_SMALL(params.template get<eFreeDir2>(), unitDir[eMom2], eps,
                       eps);
  CHECK_CLOSE_OR_SMALL(params.template get<eFreeQOverP>(), qOverP, eps, eps);
  // convenience accessors
  CHECK_CLOSE_OR_SMALL(params.fourPosition(), pos4, eps, eps);
  CHECK_CLOSE_OR_SMALL(params.position(), pos, eps, eps);
  CHECK_CLOSE_OR_SMALL(params.time(), pos4[eFreeTime], eps, eps);
  CHECK_CLOSE_OR_SMALL(params.unitDirection(), unitDir, eps, eps);
  CHECK_CLOSE_OR_SMALL(params.absoluteMomentum(), p, eps, eps);
  CHECK_CLOSE_OR_SMALL(params.transverseMomentum(),
                       p * unitDir.template head<2>().norm(), eps, eps);
  CHECK_CLOSE_OR_SMALL(params.momentum(), p * unitDir, eps, eps);
  BOOST_CHECK_EQUAL(params.charge(), q);
  // self-consistency
  CHECK_CLOSE_OR_SMALL(params.position(),
                       params.parameters().template segment<3>(eFreePos0), eps,
                       eps);
  CHECK_CLOSE_OR_SMALL(params.time(), params.template get<eFreeTime>(), eps,
                       eps);
}

}  // namespace

BOOST_AUTO_TEST_SUITE(CurvilinearTrackParameters)

BOOST_DATA_TEST_CASE(
    NeutralConstructFromAngles,
    posSymmetric* posSymmetric* posSymmetric* ts* phis* thetas* ps, x, y, z,
    time, phi, theta, p) {
  Vector4D pos4(x, y, z, time);
  Vector3D dir = makeDirectionUnitFromPhiTheta(phi, theta);

  NeutralFreeTrackParameters params(pos4, phi, theta, 1 / p);
  checkParameters(params, pos4, dir, p, 0_e);
  BOOST_CHECK(not params.covariance());

  // reassign w/ covariance
  params = NeutralFreeTrackParameters(pos4, phi, theta, 1 / p, cov);
  BOOST_CHECK(params.covariance());
  BOOST_CHECK_EQUAL(params.covariance().value(), cov);
}

BOOST_DATA_TEST_CASE(
    ChargedConstructFromAngles,
    posSymmetric* posSymmetric* posSymmetric* ts* phis* thetas* ps* qsNonZero,
    x, y, z, time, phi, theta, p, q) {
  Vector4D pos4(x, y, z, time);
  Vector3D dir = makeDirectionUnitFromPhiTheta(phi, theta);

  FreeTrackParameters params(pos4, phi, theta, q / p);
  checkParameters(params, pos4, dir, p, q);
  BOOST_CHECK(not params.covariance());

  // reassign w/ covariance
  params = FreeTrackParameters(pos4, phi, theta, q / p, cov);
  BOOST_CHECK(params.covariance());
  BOOST_CHECK_EQUAL(params.covariance().value(), cov);
}

BOOST_DATA_TEST_CASE(
    AnyConstructFromAngles,
    posSymmetric* posSymmetric* posSymmetric* ts* phis* thetas* ps* qsNonZero,
    x, y, z, time, phi, theta, p, q) {
  Vector4D pos4(x, y, z, time);
  Vector3D dir = makeDirectionUnitFromPhiTheta(phi, theta);

  AnyFreeTrackParameters params(pos4, phi, theta, p, q);
  checkParameters(params, pos4, dir, p, q);
  BOOST_CHECK(not params.covariance());

  // reassign w/ covariance
  params = AnyFreeTrackParameters(pos4, phi, theta, p, q, cov);
  BOOST_CHECK(params.covariance());
  BOOST_CHECK_EQUAL(params.covariance().value(), cov);
}

BOOST_AUTO_TEST_SUITE_END()
